require 'spec_helper'

describe RightPublish::YumRepo do

  def pkg_parts(path)
    result = {:name=>nil, :version=>nil, :arch=>nil}
    if /([A-Za-z\-+]+)-([0-9][A-Za-z0-9\-\.]*)\.(#{RightPublish::YumRepo::PKG_TYPES.join('|')})\.rpm\Z/.match(path)
      result[:name] = $1
      result[:version] = $2
      result[:arch] = $3
    end
    result
  end


  before(:each) do
    @repo_subdir = 'fake/'
    @repo = flexmock(RightPublish::RepoManager.get_repository(:yum_repo)) do |obj|
      obj.should_receive(:pull)
      obj.should_receive(:annotate)
      obj.should_receive(:push)
      obj.should_receive(:do_in_subdir).and_return {|dir,block| block.call}
      obj.should_receive(:yum_prune)
      obj.should_receive(:pkg_parts).and_return { |pkg| pkg_parts(pkg) }
      obj.should_receive(:lock).and_yield
    end

    flexmock(RightPublish::Profile) do |obj|
      obj.should_receive(:log).and_return { |str,dbg| }
    end

    flexmock(File) { |obj| obj.should_receive(:file?).and_return(true) }
  end

  [false, true].each do |should_sign|
    context "#publish#{" and sign" if should_sign}" do

      # Test the EPEL (not flat) repo profile type
      context "EPEL layout" do

        def build_expectations(pkgs, expected_dists)
          pkgs = Array(pkgs)

          pkgs.each do |pkg|
            pkg_arch = pkg_parts(pkg)[:arch]
            dest_archs = case pkg_arch
                         when RightPublish::YumRepo::SRC_ALL_ARCH
                           [RightPublish::YumRepo::SRC_ALL_PATH]
                         when RightPublish::YumRepo::BIN_ALL_ARCH
                           RightPublish::YumRepo::ARCHITECTURES
                         else
                           [pkg_arch]
                         end

            expected_dists.each do |dist|
              dist_name, release = dist.split("/")

              dest_archs.each do |arch|
                dist_path = File.join(@repo_subdir, @epel_version.to_s, dist_name.to_s, release.to_s, arch)
                @install_expectations[dist_path] ||= []
                @install_expectations[dist_path] << pkg
              end
            end
          end
          @repo.should_receive(:install_file).and_return do |pkg,dir|
            @install_expectations.should be_has_key(dir)
            @install_expectations[dir].delete(pkg)
            @install_expectations.delete(dir) if @install_expectations[dir].empty?
          end
          system_calls = @install_expectations.size
          system_calls *= 3 if @profile.config[:yum_repo][:gpg_key_id]
          system_sign_calls = (@profile.config[:yum_repo][:gpg_key_id]) ? 1 : 0
          @repo.should_receive(:system).times(system_calls).and_return(0)
          @repo.should_receive(:shellout_with_password).times(system_sign_calls).and_return(0)
        end

        before(:each) do
          @epel_dists = ["el/5", "el/6", "fc/12", "fc/13"]
          @epel_version = 1
          @install_expectations = {}
          @profile = flexmock(RightPublish::Profile) do |obj|
            obj.should_receive(:config).and_return {
              { :verbose=>false,
                :remote_storage=>{:provider=>'mockremote'},
                :local_storage=>{:provider=>'mocklocal'},
                :yum_repo=>{
                  :epel=>@epel_version,
                  :subdir=>@repo_subdir,
                  :dists=>@epel_dists,
                  :gpg_key_id=>should_sign,
                  :gpg_password=>'fakepass'}
              }
            }
          end
        end

        after(:each) do
          @install_expectations.should be_empty
        end

        # Test for publish noarch packages (multicasting)
        context "NoArch Packages" do
          it "installs a single noarch rpm file to all distribution/architectures" do
            rpm_file = 'foo-1.noarch.rpm'
            build_expectations(rpm_file,@epel_dists)
            @repo.publish(rpm_file, nil)
          end

          it "installs a single src rpm file to all dists" do
            rpm_file = 'foo-1.src.rpm'
            build_expectations(rpm_file,@epel_dists)
            @repo.publish(rpm_file, nil)
          end

          it "installs a list of noarch rpm files to all distribution/architectures" do
            rpm_files = ['foo-1.noarch.rpm', 'bar-2.noarch.rpm']
            build_expectations(rpm_files,@epel_dists)
            @repo.publish(rpm_files, nil)
          end

          it "globs and installs a directory of noarch rpm files to all distributions/architectures" do
            rpm_files = ['foo-1.noarch.rpm', 'bar-2.src.rpm']
            build_expectations(rpm_files,@epel_dists)
            flexmock(File) { |obj| obj.should_receive(:directory?).and_return(true) }
            flexmock(Dir) { |obj| obj.should_receive(:glob).and_return(rpm_files) }
            @repo.publish('somedir/', nil)
          end

          it "fails with no files in a dir" do
            flexmock(File) { |obj| obj.should_receive(:directory?).and_return(true) }
            flexmock(Dir) { |obj| obj.should_receive(:glob).and_return([]) }
            @repo.should_receive(:system).never
            expect { @repo.publish('some_dir/', nil) }.to raise_error(RuntimeError)
          end

          it "fails with files missing a rpm extenstion" do
            rpm_files = ['foo-1.noarch.rpm', 'bar-2.noarch.deb']
            @repo.should_receive(:system).never
            expect { @repo.publish(rpm_files, nil) }.to raise_error(RuntimeError)
          end

          it "fails with a mix of noarch and binary rpms" do
            rpm_files = ['foo-1.noarch.rpm', 'bar-2.x86_64.rpm']
            @repo.should_receive(:system).never
            expect { @repo.publish(rpm_files, nil) }.to raise_error(RuntimeError)
          end
        end # end NoArch context

        context "Architecture Dependant Packages" do
          before(:each) do
            @target_distribution_map = [ 'fc/12' ]
            @target_distribution_str = 'fc/12'
          end

          it "installs a single x86_64 rpm file to a dist" do
            rpm_file = 'foo-1.x86_64.rpm'
            build_expectations(rpm_file,@target_distribution_map)
            @repo.publish(rpm_file, @target_distribution_str)
          end

          it "installs a list of x86_64 rpms to a dist" do
            rpm_files = ['foo-1.x86_64.rpm', 'bar-2.x86_64.rpm']
            build_expectations(rpm_files,@target_distribution_map)
            @repo.publish(rpm_files, @target_distribution_str)
          end

          it "fails with no files in a dir" do
            flexmock(File) { |obj| obj.should_receive(:directory?).and_return(true) }
            flexmock(Dir) { |obj| obj.should_receive(:glob).and_return([]) }
            @repo.should_receive(:system).never
            expect { @repo.publish('some_dir/', @target_distribution_str) }.to raise_error(RuntimeError)
          end

          it "fails with files missing a rpm extension" do
            rpm_files = ['foo-1.x86_64.rpm', 'bar-2.i386.deb']
            @repo.should_receive(:system).never
            expect { @repo.publish(rpm_files, @target_distribution_str) }.to raise_error(RuntimeError)
          end

          it "fails with unknown architectures" do
            rpm_files = ['foo-1.ppc.rpm', 'bar-2.i386.rpm']
            @repo.should_receive(:system).never
            expect { @repo.publish(rpm_files, @target_distribution_str) }.to raise_error(RuntimeError)
          end

          it "fails when not provided with a target" do
            rpm_files = ['foo-1.x86_64.rpm', 'bar-2.i386.rpm']
            @repo.should_receive(:system).never
            expect { @repo.publish(rpm_files, nil) }.to raise_error(RuntimeError)
          end
        end # end Architecture Dependant context
      end # end EPEL context

      # Test the Flat repo profile type
      context "Flat layout" do
        def build_expectations(pkgs)
          @install_expectations = Array(pkgs)
          @repo.should_receive(:install_file).and_return do |pkg,dir|
            @install_expectations.delete(pkg)
          end
          system_calls = 1
          system_calls *= 3 if @profile.config[:yum_repo][:gpg_key_id]
          system_sign_calls = (@profile.config[:yum_repo][:gpg_key_id]) ?  1 : 0
          @repo.should_receive(:system).times(system_calls).and_return(0)
          @repo.should_receive(:shellout_with_password).times(system_sign_calls).and_return(0)
        end

        before(:each) do
          @epel_version = false
          @install_expectations = []
          @profile = flexmock(RightPublish::Profile) do |obj|
            obj.should_receive(:config).and_return {
              { :verbose=>false,
                :remote_storage=>{:provider=>'mockremote'},
                :local_storage=>{:provider=>'mocklocal'},
                :yum_repo=>{
                  :epel=>@epel_version,
                  :subdir=>@repo_subdir,
                  :gpg_key_id=>should_sign,
                  :gpg_password=>'fakepass'}
              }
            }
          end
        end

        after(:each) do
          @install_expectations.should be_empty
        end

        it "installs a single rpm file to repo" do
          rpm_file = 'foo-1.noarch.rpm'
          build_expectations(rpm_file)
          @repo.publish(rpm_file, nil)
        end

        it "installs a single src rpm file to repo" do
          rpm_file = 'foo-1.src.rpm'
          build_expectations(rpm_file)
          @repo.publish(rpm_file, nil)
        end

        it "installs a list of rpm files to repo" do
          rpm_files = ['foo-1.x86_64.rpm', 'bar-2.x86_64.rpm']
          build_expectations(rpm_files)
          @repo.publish(rpm_files, nil)
        end

        it "globs and installs a directory of rpm files to repo" do
          rpm_files = ['foo-1.noarch.rpm', 'bar-2.src.rpm']
          build_expectations(rpm_files)
          flexmock(File) { |obj| obj.should_receive(:directory?).and_return(true) }
          flexmock(Dir) { |obj| obj.should_receive(:glob).and_return(rpm_files) }
          @repo.publish('somedir/', nil)
        end

        it "succeeds with a mix of noarch and binary rpms" do
          rpm_files = ['foo-1.noarch.rpm', 'bar-2.x86_64.rpm']
          build_expectations(rpm_files)
          @repo.publish(rpm_files, nil)
        end

        it "fails with no files in a dir" do
          flexmock(File) { |obj| obj.should_receive(:directory?).and_return(true) }
          flexmock(Dir) { |obj| obj.should_receive(:glob).and_return([]) }
          @repo.should_receive(:system).never
          expect { @repo.publish('some_dir/', nil) }.to raise_error(RuntimeError)
        end

        it "fails with files missing a rpm extenstion" do
          rpm_files = ['foo-1.i386.rpm', 'bar-2.noarch.deb']
          @repo.should_receive(:system).never
          expect { @repo.publish('some_dir/', nil) }.to raise_error(RuntimeError)
        end
      end # end Flat context
    end # end #publish context
  end
end
