require 'spec_helper'

describe RightPublish::Storage do

  # Create a File entity.
  def make_file(name, expects, copy, destroy)
    file = flexmock("File:#{name}")

    # We can set some expectations on it like if we believe we're
    # going to be removed from Storage.
    destroy_calls = 0
    munge_body = false
    if expects.instance_of? Hash
      destroy_calls = (expects.has_key?(:destroy) && expects[:destroy] && 1) || 0
      munge_body = (expects.has_key?(:munge_body) && expects[:munge_body]) || false
    end

    # For simplicity use the filename as the "contents" but provide a
    # way to corrupt it to force testing content, not name matching.
    file.should_receive(:body).and_return { (munge_body && 'data_mismatch') || name }

    # Use the copy callback to insert a copy of ourselves at a different
    # path (Remote duplication).
    file.should_receive(:copy).and_return do |dir_key,dest|
      new_file = make_file(dest, nil, copy, destroy)
      copy.call(new_file)
    end

    # Use the destroy callback to remove outselves from the Files obj.
    file.should_receive(:destroy).times(destroy_calls).and_return { destroy.call(name) }
    file.should_receive(:key).and_return(name)
    file
  end

  # Create a Fog Storage Mock
  def make_storage(path, file_hash, expectations)
    # List of File mocks bound to this storage instance
    files = []

    # These procedures are used as callbacks for the File mocks to copy
    # or destroy themselves and have a side-effect on the about Array
    create_proc = Proc.new { |file| files << file }
    delete_proc = Proc.new { |file| files.delete_if { |f| f.key == file } }

    # Create the File mocks and push in the Array
    file_hash.each_pair do |name,expects|
      files << make_file(name, expects, create_proc, delete_proc )
    end

    # Fog has a weird notion of a 'Files' and 'Directory' object being
    # distinct. We care here because create is called on the 'Files'
    # rather then on the 'Directory' and it is Enumerable.
    file_list = flexmock(files)
    file_list.should_receive(:create).
      times(expectations[:create_call]).
      and_return do |args|
        new_file = make_file( args[:key], nil, create_proc, delete_proc )
        files << new_file
        new_file
      end

    # The only thing ever called on the Storage object is the
    # #get_directories interface, or at least that's all we use.
    obj = flexmock('Storage')
    obj.should_receive(:get_directories).and_return do
      # The Directory object is synonymous to an S3 bucket object
      # and is responsible for providing the Files enumerable and
      # calculating the MD5 hash of a given file.
      f = flexmock('Directory')
      f.should_receive(:compute_md5).and_return {|file| Digest::MD5.hexdigest(File.basename(file.body))}
      f.should_receive(:files).and_return(file_list)
      f.should_receive(:key).and_return(path)
      f
    end
    obj
  end

  # In order to match we just make sure that the sizes of the Files
  # objects match and that every File within A is also within B. We don't
  # have to check the reverse since the sizes implies it.
  def storage_match?(a,b)
    a_files = a.get_directories.files
    b_files = b.get_directories.files

    return false if a_files.count != b_files.count
    a_files.all? { |a_f|
      name = a_f.key
      b_files.any? { |b_f| name == b_f.key }
    }
  end

  before(:each) do
    flexmock(RightPublish::Profile) do |obj|
      obj.should_receive(:log).and_return { |str,debug|
      }
    end
  end

  context ".sync_dirs" do
    it "handles source AND destination storages that are empty" do
      src_storage = make_storage(
        'foo',
        {},
        {:create_call=>0} )

      # First copy of file1 will use create, second will be duped
      dest_storage = make_storage(
        'bar',
        {},
        {:create_call=>0} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage, :sweep=>true)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "does nothing with identical directories" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil },
        {:create_call=>0} )

      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil },
        {:create_call=>0} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "copies a new file to the destination" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil,
          'file4'=>nil },
        {:create_call=>0} )

      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil },
        {:create_call=>1} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "handles source storages that are empty" do
      src_storage = make_storage(
        'foo',
        {},
        {:create_call=>0} )

      # First copy of file1 will use create, second will be duped
      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>{:destroy=>true} },
        {:create_call=>0} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage, :sweep=>true)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "handles destination storages that are empty" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil },
        {:create_call=>0} )

      # First copy of file1 will use create, second will be duped
      dest_storage = make_storage(
        'bar',
        {},
        {:create_call=>1} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage, :sweep=>true)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "leaves an extraneous file in the destination w/o :sweep" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil },
        {:create_call=>0} )

      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil,
          'file4'=>nil },
        {:create_call=>0} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage)

      # We should NOT match here. Not sure what the usecase of this
      # is, but it's tested anyways.
      storage_match?(src_storage, dest_storage).should be_false
    end

    it "destroys an extraneous file from the destination w/ :sweep" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil },
        {:create_call=>0} )

      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil,
          'file4'=>{:destroy=>true} },
        {:create_call=>0} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage, :sweep=>true)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "destroys and copies files as needed to synchronize directories" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil,
          'file4'=>nil,
          'file6'=>nil },
        {:create_call=>0} )

      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file3'=>nil,
          'file5'=>{:destroy=>true},
          'dir2/file2'=>{:destroy=>true} },
        {:create_call=>2} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage, :sweep=>true)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "uses the remote copy facilities when making duplicates" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'dir2/file1'=>nil,
          'file4'=>nil },
        {:create_call=>0} )

      # There are two duplicates, 'file1', in the local store. To save
      # time/data we just call copy on the file that already exists in
      # the remote, notice the '0' create_calls.
      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file4'=>nil },
        {:create_call=>0} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage, :sweep=>true)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "uses the content of the file, not the name to determine equality" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'file4'=>nil },
        {:create_call=>0} )

      # :munge_body cause the MD5 sum to mismatch the file name
      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>{:munge_body=>true,:destroy=>true},
          'dir1/file2'=>nil,
          'file4'=>nil },
        {:create_call=>1} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage, :sweep=>true)
      storage_match?(src_storage, dest_storage).should be_true
    end

    it "uses create AND duplicate routines with multiple new identical files" do
      src_storage = make_storage(
        'foo',
        { 'dir1/file1'=>nil,
          'dir1/file2'=>nil,
          'dir2/dir3/file2'=>nil,
          'file2'=>nil },
        {:create_call=>0} )

      # First copy of file1 will use create, the rest will be duped
      dest_storage = make_storage(
        'bar',
        { 'dir1/file1'=>nil },
        {:create_call=>1} )

      RightPublish::Storage.sync_dirs(src_storage, dest_storage, :sweep=>true)
      storage_match?(src_storage, dest_storage).should be_true
    end
  end

  context ".guess_mime_type" do
    it 'always prefers RightPublish built-in types' do
      RightPublish::MIME_TYPES.each_pair do |mt, ext|
        RightPublish::Storage.guess_mime_type("some-file.#{ext}").should == mt
      end
    end

    other_types = {
      'html' => 'text/html',
      'txt'  => 'text/plain',
      'gz'   => 'application/x-gzip',
      'png'  => 'image/png',
      'jpg'  => 'image/jpeg'
    }

    other_types.each_pair do |ext, mt|
      it "knows about .#{ext} files" do
        RightPublish::Storage.guess_mime_type("some-file.#{ext}").should == mt
      end
    end

    it 'guesses application/octet-stream for unknown files' do
      RightPublish::Storage.guess_mime_type('some-file.xyzzywwfoobflo').should == 'application/octet-stream'
    end
  end
end
